﻿using System;
using System.Collections.Generic;

namespace NeuralNetworkExample
{
    public class NeuralNetwork
    {
        /// <summary>
        /// The correct expected result for vehicle. We consider our neural network as a binary classification network
        /// where the network can output 2 values. 0 or 1. We program the network so that if the network output is 1, 
        /// then the object must be a Vehicle.
        /// </summary>
        private const int CorrectExpectedResultForVehicle = 1;

        /// <summary>
        /// The correct expected result for vehicle. We consider our neural network as a binary classification network
        /// where the network can output 2 values. 0 or 1. We program the network so that if the network output is 0, 
        /// then the object must be a human.
        /// </summary>
        private const int CorrectExpectedResultForHuman = 0;

        /// <summary>
        /// Actually there is no strict rule about the value of learning rate. It is a tuning parameter 
        /// that will depend on the context(problem domain), which means, we will continue to experiment 
        /// and find out which learning rate works best for optimizing the weights for the specific problem.
        /// The value of learning rate should be between 0 and 1.Usually it should be a very small number,
        /// something like 0.1.                      
        /// </summary>
        private const double LearningRate = 0.1;

        /// <summary>
        /// Weight for WidthHeightRatio
        /// </summary>
        public double Weight1 { get; private set; }

        /// <summary>
        /// Weight for ReflectedLightPercentage
        /// </summary>
        public double Weight2 { get; private set; }
             
        /// <summary>
        /// Strategy
        /// ============
        /// Y = Σ (W * X) + b
        /// Detection Result = f(Y) = Activation Function
        /// 
        /// Here, we are using the Neural Network for Binary Classification. That means, The Activation function will result 
        /// a binary number 0 or 1. If result is 1, then we consider the object as a Car. If the result is 0, we consider the 
        /// object as a human.
        /// 
        /// For simplicity, we assume that, b or bias = 0.
        /// 
        /// Now, we will present our training data and calculate the Result for each set of data. 
        /// For each epoch we will calculate Result = f(Y), where Y = Σ (W * X) + b. 
        /// When presenting vehicle training Data, we expect the Result = f(Y) to be 0. So, 0 is the correct output 
        /// and 1 is the error output. When presenting human training data, we expect the Result = f(Y) to be 1. So, 1 is the
        /// correct output and 0 is the error output.
        /// </summary>
        private int DetectionFunction(double widthHeightRatio, double reflectedLightPercentage)
        {
            double weightedSum = Weight1 * widthHeightRatio + Weight2 * reflectedLightPercentage;

            weightedSum = Math.Round(weightedSum, 2);

            int detectionResult = ActivationFunction(weightedSum);

            return detectionResult;
        }

        private int ActivationFunction(double weightedSum)
        {
            /*
              This threshold value is NOT a magic number, rather it is a experimental value based on the data 
              set and the problem domain. For our problem domain, we find that, 0.5 can be a proper value that 
              can be used to do the classification. 
            */
            const double threshold = 0.50;

            if (weightedSum > threshold)
            {
                return 1;
            }
            else
            {
                return 0;
            }
        }

        private void TrainWithData(List<TrainingData> trainingData, int correctExpectedResult)
        {
            for (int i = 0; i < trainingData.Count; i++)
            {
                int detectionResult = DetectionFunction(trainingData[i].WidthHeightRatio, trainingData[i].ReflectedLightPercentage);

                if (detectionResult == correctExpectedResult)
                {
                    // The result is correct, so we do not need to change the Weights.
                }
                else
                {
                    int currentOutput = detectionResult;
                    int errorProduced = correctExpectedResult - currentOutput;

                    double changeInWeight1 = LearningRate * trainingData[i].WidthHeightRatio * errorProduced;
                    double newWeight1 = Weight1 + changeInWeight1;

                    Weight1 = newWeight1;

                    double changeInWeight2 = LearningRate * trainingData[i].ReflectedLightPercentage *
                                             errorProduced;
                    double newWeight2 = Weight2 + changeInWeight2;

                    Weight2 = newWeight2;
                }
            }
        }

        public void Learn(List<TrainingData> humanTrainingDataList, List<TrainingData> vehicleTrainingDataList)
        {
            /* We have 2 weights variable in this network. One weight is for WidthHeightRatio (Weight1) and another 
             * Weight is for for ReflectedLightPercentage (Weight2)
             * 
             * Before the training, we initialize the 2 Weigths with some random values. 
             * Over time, these weights will be changed based on training data. */

            Random random = new Random();

            Weight1 = random.NextDouble();
            Weight2 = random.NextDouble();

            Console.WriteLine("Learning started with vehicle data...");

            TrainWithData(vehicleTrainingDataList, CorrectExpectedResultForVehicle);

            Console.WriteLine("\nLearning successfully completed with vehicle Data.\n");

            Console.WriteLine("Learning started with human data...");

            TrainWithData(humanTrainingDataList, CorrectExpectedResultForHuman);

            Console.WriteLine("\nLearning successfully completed with human Data.\n");

            Console.WriteLine("Value of W1 = {0}", Weight1);

            Console.WriteLine("Value of W2 = {0}\n", Weight2);
        }

        public void Detect()
        {
            try
            {
                Console.Write("\nEnter width to height ratio:");

                double widthHeightRatio = Convert.ToDouble(Console.ReadLine());

                Console.Write("Enter percentage of reflected light:");

                double percentageOfReflectedLight = Convert.ToDouble(Console.ReadLine());

                int detectionResult = DetectionFunction(widthHeightRatio, percentageOfReflectedLight);

                string alertMessage;

                switch (detectionResult)
                {
                    case CorrectExpectedResultForVehicle:
                    {
                        alertMessage = "Vehicle Alert!";
                    }
                        break;
                    case CorrectExpectedResultForHuman:
                    {
                        alertMessage = "Pedestrian Alert!";
                    }
                        break;
                    default:
                        throw new Exception("Invalid Result");
                }

                Console.WriteLine(alertMessage);
            }
            catch (FormatException)
            {
                Console.WriteLine("\nSorry, You have not entered valid double number.");
            }
        }
    }
}